#include <cstdlib>	//srand()
#include <iostream>	//cout
#include <ctime>	//time()
#include <cstring>	//strcmp()
#include "test.h"	//testing functions
#include "RSA.h"	//GenerateKeyPair()
#include "PrimeGenerator.h"	//Generate()
#include "CMd5.h"

using std::cout;
using std::endl;
using std::cin;
using namespace std;
/*read a key from a key file*/
Key ReadKey(const char * keyfile)
{
	std::ifstream source(keyfile, std::ios::in | std::ios::binary);
	if (!source){
		cout<< "Error : Opening file \"KeyFile\" failed."<<endl;
	}
	string str_modulus , str_exponent;
	source >> str_modulus >> str_exponent;
	BigInt modulus(str_modulus);
	BigInt exponent(str_exponent);
	Key key(modulus , exponent);
	return key;
}

/*read whole text from a file*/
std::string GetWholeTextFromFile(const char * filename)
{
	std::ifstream source(filename , std::ios::in | std::ios::binary);
	if(!source){
		cout<< "Error : Opening file \"File\" failed."<<endl;
	}
	source.seekg(0, std::ios::end);
	const unsigned long int fileSize = source.tellg();
	source.seekg(0, std::ios::beg);
	char buff[10240];
	source.read(buff , fileSize);
	buff[fileSize] = '\0';
	std::string wholetext = std::string(buff);
	source.close();
	return wholetext;
}

/*generate a fingerprint using MD5 algorithm*/
std::string GenerateFingerPrintFromFile(const char * filename)
{
	std::ifstream source(filename , std::ios::in | std::ios::binary);
	if(!source){
		cout<< "Error : Opening file \"File\" failed."<<endl;
	}
	source.seekg(0, std::ios::end);
	const unsigned long int fileSize = source.tellg();
	source.seekg(0, std::ios::beg);
	char buff[10240];
	source.read(buff , fileSize);
	buff[fileSize] = '\0';

	unsigned char digest[16];
	//MD5غbuffMD5룬digest
	md5_state_t md5state;
    md5_init(&md5state);
    md5_append(&md5state, (const unsigned char *)buff, strlen(buff));
    md5_finish(&md5state, digest);
	char presentation[33];
	md5_presentation(digest , presentation);
	presentation[32] = '\0';
	std::string fingerPrint = std::string(presentation);
	source.close();
	return fingerPrint;
}

/*
	Send a messageFile M to a receiver.
	Pass messageFile'path , sender's private key path , receiver's public key path to the function.
	This function will write M's digitalSignature to a file called 'DigitalSignature.txt';
	Also write encrypted M a file called 'cypherFile.txt' using receiver's public key.
*/
void send(char * messageFile , char * sender_private_key , char * receiver_public_key)
{
	char cypherFile [] = "cypherFile.txt";//output file name
	char digitalSignatureFile [] = "DigitalSignature.txt";
	std:string fingerPrint = GenerateFingerPrintFromFile(messageFile);
	cout<<"1) File M 's fingerPrint is"<<endl<<fingerPrint<<endl;
	Key B_publickey = ReadKey(receiver_public_key);
	Key A_privatekey = ReadKey(sender_private_key);
	std::string cypherText = RSA::Encrypt(fingerPrint, A_privatekey);
	cout<<"2) FingerPrint has been encrypted using sender's private key"<<endl
		<<"3) DigitalSignatureFile generated , save as a file called 'DigitalSignature.txt'"<<endl;
	RSA::Encrypt(messageFile, cypherFile, B_publickey);
	cout<<"4) File M has been encrypted using receiver's public key."<<endl
		  <<"5) Save as a file called 'cypherFile.txt'"<<endl;
	std::ofstream dest(digitalSignatureFile, std::ios::out | std::ios::binary);
	dest<<cypherText;
	dest.flush();
	dest.close();
}


/*
	Receive a encrypted file cypherFile and a digitalSignature file From sender , 
	check whether M is completed or not , decrypt from cypherFile using receiver's private key.
	Pass paths of cypherFile , digitalSignatureFile ,sender_public_key ,receiver_private_key to this function.
	This function will print some checking results.
*/
void receive(char * cypherFile , char * digitalSignatureFile , char * sender_public_key , char * receiver_private_key)
{
	char newmessage [] = "newmessage.txt"; // output file name
	Key A_publickey = ReadKey(sender_public_key);
	Key B_privatekey = ReadKey(receiver_private_key);
	std::string cypherText = GetWholeTextFromFile(digitalSignatureFile);
	cout<<"1) DigitalSignatureFile has been decrypted using sender's public key."<<endl
		<<"2) Get FingerPrint"<<endl;
	std::string newFingerPrint = RSA::Decrypt(cypherText, A_publickey);
	cout<<newFingerPrint<<endl;
	RSA::Decrypt(cypherFile, newmessage, B_privatekey);
	cout<<"3) CypherFile has been decrypted using B's private key."<<endl
		<<"4) Get another printfinger:"<<endl;
	std::string anotherFingerPrint = GenerateFingerPrintFromFile(newmessage);
	cout<<anotherFingerPrint<<endl;
	cout<<"5) Now check two printfingers"<<endl;
	bool check = (newFingerPrint==anotherFingerPrint);
	if (check == true)
	{
		cout<<"Result : File is completed!"<<endl;
	}
	else
	{
		cout<<"Result : File is not completed,Please check!"<<endl;
	}
}

int main(int argc, char *argv[])
{
	char messageFile [] = "message.txt";
	char sender_private_key [] = "A_private.key";
	char receiver_public_key [] = "B_public.key";

	char sender_public_key [] = "A_public.key";
	char receiver_private_key [] = "B_private.key";
	char digitalSignature [] = "DigitalSignature.txt";
	char cypherFile [] = "cypherFile.txt";

	cout<<"***************Read Me***************"<<endl;
	cout<<"Please read this algotirhm's document first!"<<endl;
	cout<<"This program uses MD5 as hash function and RSA as Asymmetric Encryption Algorithm."<<endl;
	cout<<"Please check that you must have both sender and receiver's public key and private key"<<endl;
	cout<<"Please check that you must have a message file to be test."<<endl;
	cout<<"File structure must look like as following:"<<endl;
	cout<<"message.txt,A_private.key,A_public.key,B_private.key,B_public.key"<<endl;
	cout<<"*************************************"<<endl;
	std::string input;
	while(1)
	{
		cout<<"If you are sure , please input 'yes' to start this program!"<<endl;
		cin >> input;
		if (input != "yes")
			continue;
		cout<<"*****************Send****************"<<endl;
		send(messageFile , sender_private_key , receiver_public_key);
		cout<<"***************Receive***************"<<endl;
		receive(cypherFile ,digitalSignature , sender_public_key , receiver_private_key);
		break;
	}
	cout<<"Please press any key to exit!"<<endl;
	cin >> input;
	return 0;
}